from __future__ import annotations

import tempfile
from typing import Any

import joblib
import lightgbm as lgb
import numpy as np
import pytest
from factrainer.core import (
    CvModelContainer,
    PredMode,
)
from factrainer.lightgbm import (
    LgbDataset,
    LgbModel,
    LgbModelConfig,
    LgbPredictConfig,
    LgbTrainConfig,
)
from numpy import typing as npt
from numpy.testing import assert_allclose
from sklearn.metrics import r2_score
from sklearn.model_selection import KFold, train_test_split


@pytest.mark.flaky(reruns=3, reruns_delay=5, only_rerun=["HTTPError"])
def test_cv_pred_config(
    california_housing_data: tuple[
        npt.NDArray[np.number[Any]], npt.NDArray[np.number[Any]]
    ],
) -> None:
    features, target = california_housing_data
    dataset = LgbDataset(dataset=lgb.Dataset(features, label=target))
    config = LgbModelConfig.create(
        train_config=LgbTrainConfig(
            params={"objective": "regression", "seed": 1, "deterministic": True},
            callbacks=[lgb.early_stopping(100, verbose=False)],
        ),
        pred_config=LgbPredictConfig(num_iteration=2),
    )
    k_fold = KFold(n_splits=4, shuffle=True, random_state=1)
    model = CvModelContainer(config, k_fold)
    model.train(dataset)
    y_pred = model.predict(dataset)
    metric = r2_score(target, y_pred)

    assert_allclose(metric, 0.22, atol=2.5e-02)


@pytest.mark.flaky(reruns=3, reruns_delay=5, only_rerun=["HTTPError"])
def test_cv_set_pred_config_after_trainig(
    california_housing_data: tuple[
        npt.NDArray[np.number[Any]], npt.NDArray[np.number[Any]]
    ],
) -> None:
    features, target = california_housing_data
    dataset = LgbDataset(dataset=lgb.Dataset(features, label=target))
    config = LgbModelConfig.create(
        train_config=LgbTrainConfig(
            params={"objective": "regression", "seed": 1, "deterministic": True},
            callbacks=[lgb.early_stopping(100, verbose=False)],
        ),
        pred_config=LgbPredictConfig(num_iteration=2),
    )
    k_fold = KFold(n_splits=4, shuffle=True, random_state=1)
    model = CvModelContainer(config, k_fold)
    model.train(dataset)
    metric_underfit = r2_score(target, model.predict(dataset))
    model.pred_config = LgbPredictConfig(num_iteration=None)
    metric = r2_score(target, model.predict(dataset))

    assert_allclose(metric_underfit, 0.22, atol=2.5e-02)
    assert_allclose(metric, 0.84, atol=2.5e-02)


@pytest.mark.flaky(reruns=3, reruns_delay=5, only_rerun=["HTTPError"])
def test_cv_average_ensembling(
    california_housing_data: tuple[
        npt.NDArray[np.number[Any]], npt.NDArray[np.number[Any]]
    ],
) -> None:
    features, target = california_housing_data
    train_X, test_X, train_y, test_y = train_test_split(
        features,
        target,
        test_size=0.2,
        random_state=1,
    )
    train_dataset = LgbDataset(dataset=lgb.Dataset(train_X, train_y))
    test_dataset = LgbDataset(dataset=lgb.Dataset(test_X, test_y))
    config = LgbModelConfig.create(
        train_config=LgbTrainConfig(
            params={"objective": "regression", "seed": 1, "deterministic": True},
            callbacks=[lgb.early_stopping(100, verbose=False)],
        ),
    )
    k_fold = KFold(n_splits=5, shuffle=True, random_state=1)
    model = CvModelContainer(config, k_fold)
    model.train(train_dataset, n_jobs=4)
    y_pred = model.predict(test_dataset, n_jobs=4, mode=PredMode.AVG_ENSEMBLE)
    metric = r2_score(test_y, y_pred)

    assert_allclose(metric, 0.84, atol=2.5e-02)


@pytest.mark.flaky(reruns=3, reruns_delay=5, only_rerun=["HTTPError"])
def test_cv_model_picklable(
    california_housing_data: tuple[
        npt.NDArray[np.number[Any]], npt.NDArray[np.number[Any]]
    ],
) -> None:
    features, target = california_housing_data
    dataset = LgbDataset(dataset=lgb.Dataset(features, label=target))
    config = LgbModelConfig.create(
        train_config=LgbTrainConfig(
            params={"objective": "regression", "seed": 1, "deterministic": True},
            callbacks=[lgb.early_stopping(100, verbose=False)],
        ),
    )
    k_fold = KFold(n_splits=4, shuffle=True, random_state=1)
    model = CvModelContainer(config, k_fold)
    model.train(dataset, n_jobs=4)
    with tempfile.NamedTemporaryFile() as fp:
        joblib.dump(model, fp.name)
        loaded_model: CvModelContainer[
            LgbDataset, LgbModel, LgbTrainConfig, LgbPredictConfig
        ] = joblib.load(fp.name)
    y_pred = loaded_model.predict(dataset, n_jobs=4)
    metric = r2_score(target, y_pred)

    assert_allclose(metric, 0.84, atol=2.5e-02)



from __future__ import annotations

import lightgbm as lgb
import numpy as np
from factrainer.core.cv.dataset import (
    IndexedDataset,
    IndexedDatasets,
    SplittedDataset,
    SplittedDatasets,
    SplittedDatasetsIndices,
)
from factrainer.lightgbm.dataset.dataset import LgbDataset
from numpy.testing import assert_array_equal
from sklearn.model_selection import KFold


class TestSplittedDatasets:
    def test_create_from_k_fold(self) -> None:
        k_fold = KFold(n_splits=3, shuffle=True, random_state=1)
        dataset = LgbDataset(dataset=lgb.Dataset(np.array([[1, 2], [3, 4], [5, 6]])))
        expected = SplittedDatasets[LgbDataset](
            datasets=[
                SplittedDataset(
                    train=IndexedDataset(
                        index=[1, 2],
                        data=LgbDataset(
                            dataset=lgb.Dataset(np.array([[3, 4], [5, 6]]))
                        ),
                    ),
                    val=IndexedDataset(
                        index=[0],
                        data=LgbDataset(dataset=lgb.Dataset(np.array([[1, 2]]))),
                    ),
                    test=IndexedDataset(
                        index=[0],
                        data=LgbDataset(dataset=lgb.Dataset(np.array([[1, 2]]))),
                    ),
                ),
                SplittedDataset(
                    train=IndexedDataset(
                        index=[0, 1],
                        data=LgbDataset(
                            dataset=lgb.Dataset(np.array([[1, 2], [3, 4]]))
                        ),
                    ),
                    val=IndexedDataset(
                        index=[2],
                        data=LgbDataset(dataset=lgb.Dataset(np.array([[5, 6]]))),
                    ),
                    test=IndexedDataset(
                        index=[2],
                        data=LgbDataset(dataset=lgb.Dataset(np.array([[5, 6]]))),
                    ),
                ),
                SplittedDataset(
                    train=IndexedDataset(
                        index=[0, 2],
                        data=LgbDataset(
                            dataset=lgb.Dataset(np.array([[1, 2], [5, 6]]))
                        ),
                    ),
                    val=IndexedDataset(
                        index=[1],
                        data=LgbDataset(dataset=lgb.Dataset(np.array([[3, 4]]))),
                    ),
                    test=IndexedDataset(
                        index=[1],
                        data=LgbDataset(dataset=lgb.Dataset(np.array([[3, 4]]))),
                    ),
                ),
            ]
        )

        actual = SplittedDatasets[LgbDataset].create(dataset, k_fold)

        for actual_dataset, expected_dataset in zip(actual.datasets, expected.datasets):
            assert_array_equal(actual_dataset.train.index, expected_dataset.train.index)
            assert_array_equal(actual_dataset.val.index, expected_dataset.val.index)  # type: ignore
            assert_array_equal(actual_dataset.test.index, expected_dataset.test.index)

            assert_array_equal(
                actual_dataset.train.data.dataset.data,
                expected_dataset.train.data.dataset.data,
            )
            assert_array_equal(
                actual_dataset.val.data.dataset.data,  # type: ignore
                expected_dataset.val.data.dataset.data,  # type: ignore
            )
            assert_array_equal(
                actual_dataset.test.data.dataset.data,
                expected_dataset.test.data.dataset.data,
            )

    def test_create_from_indices(self) -> None:
        k_fold = SplittedDatasetsIndices(
            train=[[1, 2], [0, 1], [0, 2]], val=[[0], [2], [1]], test=[[0], [2], [1]]
        )
        dataset = LgbDataset(dataset=lgb.Dataset(np.array([[1, 2], [3, 4], [5, 6]])))
        expected = SplittedDatasets[LgbDataset](
            datasets=[
                SplittedDataset(
                    train=IndexedDataset(
                        index=[1, 2],
                        data=LgbDataset(
                            dataset=lgb.Dataset(np.array([[3, 4], [5, 6]]))
                        ),
                    ),
                    val=IndexedDataset(
                        index=[0],
                        data=LgbDataset(dataset=lgb.Dataset(np.array([[1, 2]]))),
                    ),
                    test=IndexedDataset(
                        index=[0],
                        data=LgbDataset(dataset=lgb.Dataset(np.array([[1, 2]]))),
                    ),
                ),
                SplittedDataset(
                    train=IndexedDataset(
                        index=[0, 1],
                        data=LgbDataset(
                            dataset=lgb.Dataset(np.array([[1, 2], [3, 4]]))
                        ),
                    ),
                    val=IndexedDataset(
                        index=[2],
                        data=LgbDataset(dataset=lgb.Dataset(np.array([[5, 6]]))),
                    ),
                    test=IndexedDataset(
                        index=[2],
                        data=LgbDataset(dataset=lgb.Dataset(np.array([[5, 6]]))),
                    ),
                ),
                SplittedDataset(
                    train=IndexedDataset(
                        index=[0, 2],
                        data=LgbDataset(
                            dataset=lgb.Dataset(np.array([[1, 2], [5, 6]]))
                        ),
                    ),
                    val=IndexedDataset(
                        index=[1],
                        data=LgbDataset(dataset=lgb.Dataset(np.array([[3, 4]]))),
                    ),
                    test=IndexedDataset(
                        index=[1],
                        data=LgbDataset(dataset=lgb.Dataset(np.array([[3, 4]]))),
                    ),
                ),
            ]
        )

        actual = SplittedDatasets[LgbDataset].create(dataset, k_fold)

        for actual_dataset, expected_dataset in zip(actual.datasets, expected.datasets):
            assert_array_equal(actual_dataset.train.index, expected_dataset.train.index)
            assert_array_equal(actual_dataset.val.index, expected_dataset.val.index)  # type: ignore
            assert_array_equal(actual_dataset.test.index, expected_dataset.test.index)

            assert_array_equal(
                actual_dataset.train.data.dataset.data,
                expected_dataset.train.data.dataset.data,
            )
            assert_array_equal(
                actual_dataset.val.data.dataset.data,  # type: ignore
                expected_dataset.val.data.dataset.data,  # type: ignore
            )
            assert_array_equal(
                actual_dataset.test.data.dataset.data,
                expected_dataset.test.data.dataset.data,
            )


class TestIndexedDatasets:
    def test_create(self) -> None:
        dataset = LgbDataset(dataset=lgb.Dataset(np.array([[1, 2], [3, 4], [5, 6]])))
        indices = [[0, 1], [1, 2], [2, 0]]
        expected = IndexedDatasets[LgbDataset](
            datasets=[
                IndexedDataset(
                    index=[0, 1],
                    data=LgbDataset(dataset=lgb.Dataset(np.array([[1, 2], [3, 4]]))),
                ),
                IndexedDataset(
                    index=[1, 2],
                    data=LgbDataset(dataset=lgb.Dataset(np.array([[3, 4], [5, 6]]))),
                ),
                IndexedDataset(
                    index=[2, 0],
                    data=LgbDataset(dataset=lgb.Dataset(np.array([[5, 6], [1, 2]]))),
                ),
            ]
        )
        actual = IndexedDatasets[LgbDataset].create(dataset, indices)
        for actual_dataset, expected_dataset in zip(actual.datasets, expected.datasets):
            assert_array_equal(actual_dataset.index, expected_dataset.index)
            assert_array_equal(
                actual_dataset.data.dataset.data,
                expected_dataset.data.dataset.data,
            )



from typing import Any
from unittest.mock import MagicMock, patch

import lightgbm as lgb
import numpy as np
from factrainer.lightgbm.config import (
    LgbLearner,
    LgbModelConfig,
    LgbPredictConfig,
    LgbPredictor,
    LgbTrainConfig,
)
from factrainer.lightgbm.dataset.dataset import LgbDataset
from factrainer.lightgbm.raw_model import LgbModel
from lightgbm.basic import _LGBM_TrainDataType
from numpy import typing as npt


@patch("factrainer.lightgbm.config.LgbPredictor", spec=LgbPredictor)
@patch("factrainer.lightgbm.config.LgbLearner", spec=LgbLearner)
class TestLgbModelConfig:
    @patch("factrainer.lightgbm.config.LgbPredictConfig", spec=LgbPredictConfig)
    def test_create_lgb_model_config_with_default(
        self,
        pred_config: MagicMock,
        learner: MagicMock,
        predictor: MagicMock,
    ) -> None:
        train_config = MagicMock(spec=LgbTrainConfig)
        expected = LgbModelConfig(
            learner=learner.return_value,
            predictor=predictor.return_value,
            train_config=train_config,
            pred_config=pred_config.return_value,
        )
        actual = LgbModelConfig.create(train_config)
        assert actual == expected

    def test_create_lgb_model_config(
        self, learner: MagicMock, predictor: MagicMock
    ) -> None:
        train_config = MagicMock(spec=LgbTrainConfig)
        pred_config = MagicMock(spec=LgbPredictConfig)
        expected = LgbModelConfig(
            learner=learner.return_value,
            predictor=predictor.return_value,
            train_config=train_config,
            pred_config=pred_config,
        )
        actual = LgbModelConfig.create(train_config, pred_config)
        assert actual == expected


@patch("lightgbm.train")
class TestLgbLearner:
    def test_train_with_val(self, lgb_train: MagicMock) -> None:
        train_lgb_dataset = MagicMock(spec=lgb.Dataset)
        train_dataset = LgbDataset(dataset=train_lgb_dataset)
        val_lgb_dataset = MagicMock(spec=lgb.Dataset)
        val_dataset = LgbDataset(dataset=val_lgb_dataset)
        config = LgbTrainConfig(params={"foo": "bar"})
        lgb_train.return_value = MagicMock(spec=lgb.Booster)
        expected = LgbModel(model=lgb_train.return_value)
        sut = LgbLearner()

        actual = sut.train(train_dataset, val_dataset, config)

        assert actual == expected
        lgb_train.assert_called_once_with(
            params={"foo": "bar"},
            num_boost_round=100,
            valid_names=None,
            feval=None,
            init_model=None,
            keep_training_booster=False,
            callbacks=None,
            train_set=train_lgb_dataset,
            valid_sets=[val_lgb_dataset],
        )

    def test_train_without_val(self, lgb_train: MagicMock) -> None:
        train_lgb_dataset = MagicMock(spec=lgb.Dataset)
        train_dataset = LgbDataset(dataset=train_lgb_dataset)
        val_dataset = None
        config = LgbTrainConfig(params={"foo": "bar"})

        lgb_train.return_value = MagicMock(spec=lgb.Booster)
        expected = LgbModel(model=lgb_train.return_value)
        sut = LgbLearner()

        actual = sut.train(train_dataset, val_dataset, config)

        assert actual == expected
        lgb_train.assert_called_once_with(
            params={"foo": "bar"},
            num_boost_round=100,
            valid_names=None,
            feval=None,
            init_model=None,
            keep_training_booster=False,
            callbacks=None,
            train_set=train_lgb_dataset,
            valid_sets=None,
        )

    def test_train_without_default(self, lgb_train: MagicMock) -> None:
        def custom_metric(
            preds: npt.NDArray[Any], eval_data: lgb.Dataset
        ) -> tuple[str, float, bool]:
            return "custom_metric", 0.0, True

        def custom_callback() -> None: ...

        init_model = MagicMock(spec=lgb.Booster)
        callbacks = [custom_callback]
        train_lgb_dataset = MagicMock(spec=lgb.Dataset)
        train_dataset = LgbDataset(dataset=train_lgb_dataset)
        val_dataset = None
        config = LgbTrainConfig(
            params={"foo": "bar"},
            num_boost_round=1000,
            valid_names=["foo", "bar"],
            feval=custom_metric,
            init_model=init_model,
            keep_training_booster=True,
            callbacks=callbacks,
        )

        lgb_train.return_value = MagicMock(spec=lgb.Booster)
        expected = LgbModel(model=lgb_train.return_value)
        sut = LgbLearner()

        actual = sut.train(train_dataset, val_dataset, config)

        assert actual == expected
        lgb_train.assert_called_once_with(
            params={"foo": "bar"},
            num_boost_round=1000,
            valid_names=["foo", "bar"],
            feval=custom_metric,
            init_model=init_model,
            keep_training_booster=True,
            callbacks=callbacks,
            train_set=train_lgb_dataset,
            valid_sets=None,
        )


class TestLgbPredictor:
    def test_predict_with_default_config(self) -> None:
        data = MagicMock(spec=_LGBM_TrainDataType)
        dataset = LgbDataset(dataset=lgb.Dataset(data))
        raw_model = MagicMock(spec=lgb.Booster)
        raw_model.predict.return_value = MagicMock(spec=np.ndarray)
        model = LgbModel(model=raw_model)
        sut = LgbPredictor()

        actual = sut.predict(dataset, model, LgbPredictConfig())

        assert actual == raw_model.predict.return_value
        raw_model.predict.assert_called_once_with(
            data=data,
            start_iteration=0,
            num_iteration=None,
            raw_score=False,
            pred_leaf=False,
            pred_contrib=False,
            data_has_header=False,
            validate_features=False,
        )

    def test_predict_with_config(self) -> None:
        data = MagicMock(spec=_LGBM_TrainDataType)
        dataset = LgbDataset(dataset=lgb.Dataset(data))
        raw_model = MagicMock(spec=lgb.Booster)
        raw_model.predict.return_value = MagicMock(spec=np.ndarray)
        model = LgbModel(model=raw_model)
        config = LgbPredictConfig(
            start_iteration=10,
            num_iteration=1000,
            raw_score=True,
            pred_leaf=True,
            pred_contrib=True,
            data_has_header=True,
            validate_features=True,
        )
        sut = LgbPredictor()

        actual = sut.predict(dataset, model, config)

        assert actual == raw_model.predict.return_value
        raw_model.predict.assert_called_once_with(
            data=data,
            start_iteration=10,
            num_iteration=1000,
            raw_score=True,
            pred_leaf=True,
            pred_contrib=True,
            data_has_header=True,
            validate_features=True,
        )

    def test_predict_with_custom_kwargs_config(self) -> None:
        data = MagicMock(spec=_LGBM_TrainDataType)
        dataset = LgbDataset(dataset=lgb.Dataset(data))
        raw_model = MagicMock(spec=lgb.Booster)
        raw_model.predict.return_value = MagicMock(spec=np.ndarray)
        model = LgbModel(model=raw_model)
        config = LgbPredictConfig(
            start_iteration=10,
            num_iteration=1000,
            raw_score=True,
            pred_leaf=True,
            pred_contrib=True,
            data_has_header=True,
            validate_features=True,
            custom_kwargs="foo",
        )
        sut = LgbPredictor()

        actual = sut.predict(dataset, model, config)

        assert actual == raw_model.predict.return_value
        raw_model.predict.assert_called_once_with(
            data=data,
            start_iteration=10,
            num_iteration=1000,
            raw_score=True,
            pred_leaf=True,
            pred_contrib=True,
            data_has_header=True,
            validate_features=True,
            custom_kwargs="foo",
        )



from unittest.mock import MagicMock, PropertyMock, patch

from factrainer.base.config import (
    BaseLearner,
    BaseMlModelConfig,
    BasePredictConfig,
    BasePredictor,
    BaseTrainConfig,
)
from factrainer.base.dataset import IndexableDataset
from factrainer.base.raw_model import RawModel
from factrainer.core.cv.config import PredMode
from factrainer.core.cv.dataset import SplittedDatasetsIndices
from factrainer.core.cv.model_container import CvModelContainer
from sklearn.model_selection._split import _BaseKFold


@patch("factrainer.core.cv.model_container.CvLearner", spec=BaseLearner)
@patch("factrainer.core.cv.dataset.SplittedDatasets.create")
class TestTrain:
    def test_train(
        self,
        create: MagicMock,
        CvLearner: MagicMock,
    ) -> None:
        dataset = MagicMock(spec=IndexableDataset).return_value
        config = MagicMock(spec=BaseMlModelConfig).return_value
        k_fold = MagicMock(spec=_BaseKFold | SplittedDatasetsIndices).return_value
        sut = CvModelContainer[
            IndexableDataset, RawModel, BaseTrainConfig, BasePredictConfig
        ](config, k_fold)

        sut.train(dataset)

        create.assert_called_once_with(dataset, k_fold)
        CvLearner.assert_called_once_with(config.learner)
        CvLearner.return_value.train.assert_called_once_with(
            create.return_value.train,
            create.return_value.val,
            config.train_config,
            None,
        )
        assert sut.cv_indices == create.return_value.indices
        assert sut.raw_model == CvLearner.return_value.train.return_value

    def test_parallel(
        self,
        create: MagicMock,
        CvLearner: MagicMock,
    ) -> None:
        dataset = MagicMock(spec=IndexableDataset).return_value
        config = MagicMock(spec=BaseMlModelConfig).return_value
        k_fold = MagicMock(spec=_BaseKFold | SplittedDatasetsIndices).return_value
        sut = CvModelContainer[
            IndexableDataset, RawModel, BaseTrainConfig, BasePredictConfig
        ](config, k_fold)

        sut.train(dataset, n_jobs=2)

        create.assert_called_once_with(dataset, k_fold)
        CvLearner.assert_called_once_with(config.learner)
        CvLearner.return_value.train.assert_called_once_with(
            create.return_value.train,
            create.return_value.val,
            config.train_config,
            2,
        )
        assert sut.cv_indices == create.return_value.indices
        assert sut.raw_model == CvLearner.return_value.train.return_value


@patch.object(
    CvModelContainer,
    "raw_model",
    new_callable=PropertyMock,
)
class TestPredict:
    @patch.object(
        CvModelContainer,
        "cv_indices",
        new_callable=PropertyMock,
    )
    @patch("factrainer.core.cv.dataset.IndexedDatasets.create")
    @patch("factrainer.core.cv.model_container.OutOfFoldPredictor", spec=BasePredictor)
    def test_oof_predict(
        self,
        MockOofPredictor: MagicMock,
        create: MagicMock,
        cv_indices: MagicMock,
        raw_model: MagicMock,
    ) -> None:
        dataset = MagicMock(spec=IndexableDataset).return_value
        config = MagicMock(spec=BaseMlModelConfig).return_value
        k_fold = MagicMock(spec=_BaseKFold | SplittedDatasetsIndices).return_value
        sut = CvModelContainer[
            IndexableDataset, RawModel, BaseTrainConfig, BasePredictConfig
        ](config, k_fold)

        actual = sut.predict(dataset)

        create.assert_called_once_with(dataset, cv_indices.return_value.test)
        MockOofPredictor.assert_called_once_with(config.predictor)
        MockOofPredictor.return_value.predict.assert_called_once_with(
            create.return_value,
            raw_model.return_value,
            config.pred_config,
            None,
        )
        assert actual == MockOofPredictor.return_value.predict.return_value

    @patch(
        "factrainer.core.cv.model_container.AverageEnsemblePredictor",
        spec=BasePredictor,
    )
    def test_average_ensemble_predict(
        self,
        MockAvgEnsemblePredictor: MagicMock,
        raw_model: MagicMock,
    ) -> None:
        dataset = MagicMock(spec=IndexableDataset).return_value
        config = MagicMock(spec=BaseMlModelConfig).return_value
        k_fold = MagicMock(spec=_BaseKFold | SplittedDatasetsIndices).return_value
        sut = CvModelContainer[
            IndexableDataset, RawModel, BaseTrainConfig, BasePredictConfig
        ](config, k_fold)

        actual = sut.predict(dataset, mode=PredMode.AVG_ENSEMBLE)

        MockAvgEnsemblePredictor.assert_called_once_with(config.predictor)
        MockAvgEnsemblePredictor.return_value.predict.assert_called_once_with(
            dataset,
            raw_model.return_value,
            config.pred_config,
            None,
        )
        assert actual == MockAvgEnsemblePredictor.return_value.predict.return_value

    @patch.object(
        CvModelContainer,
        "cv_indices",
        new_callable=PropertyMock,
    )
    @patch("factrainer.core.cv.dataset.IndexedDatasets.create")
    @patch("factrainer.core.cv.model_container.OutOfFoldPredictor", spec=BasePredictor)
    def test_parallel(
        self,
        MockOofPredictor: MagicMock,
        create: MagicMock,
        cv_indices: MagicMock,
        raw_model: MagicMock,
    ) -> None:
        dataset = MagicMock(spec=IndexableDataset).return_value
        config = MagicMock(spec=BaseMlModelConfig).return_value
        k_fold = MagicMock(spec=_BaseKFold | SplittedDatasetsIndices).return_value
        sut = CvModelContainer[
            IndexableDataset, RawModel, BaseTrainConfig, BasePredictConfig
        ](config, k_fold)

        actual = sut.predict(dataset, 2)

        create.assert_called_once_with(dataset, cv_indices.return_value.test)
        MockOofPredictor.assert_called_once_with(config.predictor)
        MockOofPredictor.return_value.predict.assert_called_once_with(
            create.return_value,
            raw_model.return_value,
            config.pred_config,
            2,
        )
        assert actual == MockOofPredictor.return_value.predict.return_value


class TestModelConfig:
    def test_set_train_config(self) -> None:
        train_config = MagicMock(spec=BaseTrainConfig).return_value
        model_config = MagicMock(spec=BaseMlModelConfig).return_value
        model_config.train_config = train_config
        new_train_config = MagicMock(spec=BaseTrainConfig).return_value
        sut = CvModelContainer[
            MagicMock, MagicMock, BaseTrainConfig, BasePredictConfig
        ](model_config, MagicMock())

        sut.train_config = new_train_config

        assert sut.train_config == new_train_config
        assert model_config.train_config != new_train_config

    def test_set_pred_config(self) -> None:
        pred_config = MagicMock(spec=BasePredictConfig).return_value
        model_config = MagicMock(spec=BaseMlModelConfig).return_value
        model_config.pred_config = pred_config
        new_pred_config = MagicMock(spec=BasePredictConfig).return_value
        sut = CvModelContainer[
            MagicMock, MagicMock, BaseTrainConfig, BasePredictConfig
        ](model_config, MagicMock())

        sut.pred_config = new_pred_config

        assert sut.pred_config == new_pred_config
        assert model_config.pred_config != new_pred_config



from unittest.mock import MagicMock, patch

import numpy as np
import pandas as pd
import polars as pl
from factrainer.sklearn.config import (
    SklearnLearner,
    SklearnModelConfig,
    SklearnPredictConfig,
    SklearnPredictMethod,
    SklearnPredictor,
    SklearnTrainConfig,
)
from factrainer.sklearn.dataset.dataset import SklearnDataset
from factrainer.sklearn.raw_model import SklearnModel
from sklearn.linear_model import LinearRegression, LogisticRegression


@patch("factrainer.sklearn.config.SklearnPredictor", spec=SklearnPredictor)
@patch("factrainer.sklearn.config.SklearnLearner", spec=SklearnLearner)
class TestSklearnModelConfig:
    @patch("factrainer.sklearn.config.SklearnPredictConfig", spec=SklearnPredictConfig)
    def test_create_sklearn_model_config_with_default(
        self,
        pred_config: MagicMock,
        learner: MagicMock,
        predictor: MagicMock,
    ) -> None:
        train_config = MagicMock(spec=SklearnTrainConfig)
        expected = SklearnModelConfig(
            learner=learner.return_value,
            predictor=predictor.return_value,
            train_config=train_config,
            pred_config=pred_config.return_value,
        )
        actual = SklearnModelConfig.create(train_config)
        assert actual == expected

    def test_create_sklearn_model_config(
        self, learner: MagicMock, predictor: MagicMock
    ) -> None:
        train_config = MagicMock(spec=SklearnTrainConfig)
        pred_config = MagicMock(spec=SklearnPredictConfig)
        expected = SklearnModelConfig(
            learner=learner.return_value,
            predictor=predictor.return_value,
            train_config=train_config,
            pred_config=pred_config,
        )
        actual = SklearnModelConfig.create(train_config, pred_config)
        assert actual == expected


class TestSklearnLearner:
    def test_train(self) -> None:
        dataset = SklearnDataset(X=np.array([[1, 2], [3, 4.5]]), y=np.array([0, 1]))
        config = SklearnTrainConfig(estimator=LinearRegression())

        actual = SklearnLearner().train(dataset, None, config)

        assert hasattr(actual.estimator, "intercept_")
        assert hasattr(actual.estimator, "coef_")

    def test_pandas(self) -> None:
        dataset = SklearnDataset(
            X=pd.DataFrame([[1, 2], [3, 4.5]]), y=pd.Series([0, 1])
        )
        config = SklearnTrainConfig(estimator=LinearRegression())

        actual = SklearnLearner().train(dataset, None, config)

        assert hasattr(actual.estimator, "intercept_")
        assert hasattr(actual.estimator, "coef_")

    def test_polars(self) -> None:
        dataset = SklearnDataset(
            X=pl.DataFrame({"a": [1, 2], "b": [3.0, 4.5]}), y=pl.Series([0, 1])
        )
        config = SklearnTrainConfig(estimator=LinearRegression())

        actual = SklearnLearner().train(dataset, None, config)

        assert hasattr(actual.estimator, "intercept_")
        assert hasattr(actual.estimator, "coef_")


class TestSklearnPredictor:
    @patch("sklearn.linear_model.LinearRegression.predict")
    def test_regression(self, predict: MagicMock) -> None:
        estimator = LinearRegression()
        X, y = np.array([[1, 2], [3, 4.5]]), np.array([0, 1])
        dataset = SklearnDataset(X=X, y=y)
        model = SklearnModel(estimator=estimator)
        config = SklearnPredictConfig()
        expected = predict.return_value

        actual = SklearnPredictor().predict(dataset, model, config)

        assert actual == expected
        predict.assert_called_once_with(X)

    class TestClassification:
        @patch("sklearn.linear_model.LogisticRegression.predict_proba")
        def test_auto(self, predict_proba: MagicMock) -> None:
            estimator = LogisticRegression()
            X, y = np.array([[1, 2], [3, 4.5]]), np.array([0, 1])
            dataset = SklearnDataset(X=X, y=y)
            model = SklearnModel(estimator=estimator)
            config = SklearnPredictConfig()
            expected = predict_proba.return_value

            actual = SklearnPredictor().predict(dataset, model, config)

            assert actual == expected
            predict_proba.assert_called_once_with(X)

        @patch("sklearn.linear_model.LogisticRegression.predict")
        def test_predict(self, predict: MagicMock) -> None:
            estimator = LogisticRegression()
            X, y = np.array([[1, 2], [3, 4.5]]), np.array([0, 1])
            dataset = SklearnDataset(X=X, y=y)
            model = SklearnModel(estimator=estimator)
            config = SklearnPredictConfig(predict_method=SklearnPredictMethod.PREDICT)
            expected = predict.return_value

            actual = SklearnPredictor().predict(dataset, model, config)

            assert actual == expected
            predict.assert_called_once_with(X)

        @patch("sklearn.linear_model.LogisticRegression.predict_proba")
        def test_predict_proba(self, predict_proba: MagicMock) -> None:
            estimator = LogisticRegression()
            X, y = np.array([[1, 2], [3, 4.5]]), np.array([0, 1])
            dataset = SklearnDataset(X=X, y=y)
            model = SklearnModel(estimator=estimator)
            config = SklearnPredictConfig(
                predict_method=SklearnPredictMethod.PREDICT_PROBA
            )
            expected = predict_proba.return_value

            actual = SklearnPredictor().predict(dataset, model, config)

            assert actual == expected
            predict_proba.assert_called_once_with(X)



from __future__ import annotations

from typing import Any

import numpy as np
import pytest
from numpy import typing as npt
from sklearn.datasets import (
    fetch_california_housing,
    load_breast_cancer,
    load_iris,
)


@pytest.fixture(scope="package")
def california_housing_data() -> tuple[
    npt.NDArray[np.number[Any]], npt.NDArray[np.number[Any]]
]:
    data = fetch_california_housing()
    return data["data"], data["target"]  # type: ignore


@pytest.fixture(scope="package")
def iris_data() -> tuple[npt.NDArray[np.number[Any]], npt.NDArray[np.number[Any]]]:
    return load_iris(return_X_y=True)


@pytest.fixture(scope="package")
def breast_cancer_data() -> tuple[
    npt.NDArray[np.number[Any]], npt.NDArray[np.number[Any]]
]:
    return load_breast_cancer(return_X_y=True)



from abc import ABC, abstractmethod
from typing import Any

from factrainer.base.config import BasePredictConfig, BaseTrainConfig
from factrainer.base.dataset import BaseDataset, Prediction
from factrainer.base.raw_model import RawModel


class BaseModelContainer[
    T: BaseDataset,
    U: RawModel,
    V: BaseTrainConfig,
    W: BasePredictConfig,
](ABC):
    @abstractmethod
    def train(self, train_dataset: T, *args: Any, **kwargs: Any) -> None:
        raise NotImplementedError

    @abstractmethod
    def predict(self, pred_dataset: T, *args: Any, **kwargs: Any) -> Prediction:
        raise NotImplementedError

    @property
    @abstractmethod
    def raw_model(self) -> U:
        raise NotImplementedError

    @property
    @abstractmethod
    def train_config(self) -> V:
        raise NotImplementedError

    @train_config.setter
    @abstractmethod
    def train_config(self, config: V) -> None:
        raise NotImplementedError

    @property
    @abstractmethod
    def pred_config(self) -> W:
        raise NotImplementedError

    @pred_config.setter
    @abstractmethod
    def pred_config(self, config: W) -> None:
        raise NotImplementedError



from collections.abc import Generator

import numpy as np
from factrainer.base.dataset import IndexableDataset, RowIndex, Rows
from sklearn.model_selection._split import _BaseKFold

import lightgbm as lgb

from .slicer import LgbDatasetSlicer
from .types import IsPdDataFrame


class LgbDataset(IndexableDataset):
    dataset: lgb.Dataset

    def k_fold_split(
        self, k_fold: _BaseKFold
    ) -> Generator[tuple[RowIndex, RowIndex], None, None]:
        if isinstance(self.dataset.data, np.ndarray) or IsPdDataFrame().is_instance(
            self.dataset.data
        ):
            for train_index, val_index in k_fold.split(self.dataset.data):
                yield train_index.tolist(), val_index.tolist()
        else:
            raise NotImplementedError

    def __getitem__(self, index: Rows) -> "LgbDataset":
        match index:
            case int():
                return LgbDataset(
                    dataset=LgbDatasetSlicer(self.dataset.reference).slice(
                        self.dataset, [index]
                    )
                )
            case list():
                return LgbDataset(
                    dataset=LgbDatasetSlicer(self.dataset.reference).slice(
                        self.dataset, index
                    )
                )
            case slice():
                raise NotImplementedError
            case _:
                raise TypeError



from unittest.mock import MagicMock, PropertyMock, patch

from factrainer.base.config import BaseMlModelConfig, BasePredictConfig, BaseTrainConfig
from factrainer.base.dataset import BaseDataset
from factrainer.base.raw_model import RawModel
from factrainer.core.single import SingleModelContainer


class TestTrain:
    def test_without_val(self) -> None:
        train_dataset = MagicMock(spec=BaseDataset).return_value
        config = MagicMock(spec=BaseMlModelConfig).return_value
        sut = SingleModelContainer[
            BaseDataset, RawModel, BaseTrainConfig, BasePredictConfig
        ](config)

        sut.train(train_dataset)

        assert sut.raw_model == config.learner.train.return_value
        config.learner.train.assert_called_once_with(
            train_dataset, None, config.train_config
        )

    def test_with_val(self) -> None:
        train_dataset = MagicMock(spec=BaseDataset).return_value
        val_dataset = MagicMock(spec=BaseDataset).return_value
        config = MagicMock(spec=BaseMlModelConfig).return_value
        sut = SingleModelContainer[
            BaseDataset, RawModel, BaseTrainConfig, BasePredictConfig
        ](config)

        sut.train(train_dataset, val_dataset)

        assert sut.raw_model == config.learner.train.return_value
        config.learner.train.assert_called_once_with(
            train_dataset, val_dataset, config.train_config
        )


@patch.object(
    SingleModelContainer,
    "raw_model",
    new_callable=PropertyMock,
)
def test_predict(raw_model: MagicMock) -> None:
    pred_dataset = MagicMock(spec=BaseDataset).return_value
    config = MagicMock(spec=BaseMlModelConfig).return_value
    expected = config.predictor.predict.return_value
    sut = SingleModelContainer[
        BaseDataset, RawModel, BaseTrainConfig, BasePredictConfig
    ](config)

    actual = sut.predict(pred_dataset)

    assert actual == expected
    config.predictor.predict.assert_called_once_with(
        pred_dataset, raw_model.return_value, config.pred_config
    )


class TestModelConfig:
    def test_set_train_config(self) -> None:
        train_config = MagicMock(spec=BaseTrainConfig).return_value
        model_config = MagicMock(spec=BaseMlModelConfig).return_value
        model_config.train_config = train_config
        new_train_config = MagicMock(spec=BaseTrainConfig).return_value
        sut = SingleModelContainer[
            MagicMock, MagicMock, BaseTrainConfig, BasePredictConfig
        ](model_config)

        sut.train_config = new_train_config

        assert sut.train_config == new_train_config
        assert model_config.train_config != new_train_config

    def test_set_pred_config(self) -> None:
        pred_config = MagicMock(spec=BasePredictConfig).return_value
        model_config = MagicMock(spec=BaseMlModelConfig).return_value
        model_config.pred_config = pred_config
        new_pred_config = MagicMock(spec=BasePredictConfig).return_value
        sut = SingleModelContainer[
            MagicMock, MagicMock, BaseTrainConfig, BasePredictConfig
        ](model_config)

        sut.pred_config = new_pred_config

        assert sut.pred_config == new_pred_config
        assert model_config.pred_config != new_pred_config



from typing import cast

import numpy as np
import pandas as pd
from factrainer.lightgbm.dataset.slicer import LgbDataSlicer
from numpy.testing import assert_array_equal
from pandas.testing import assert_frame_equal


def test_numpy_array() -> None:
    data = np.array([[1, 2], [3, 4], [5, 6]])
    expected = np.array([[5, 6], [1, 2]])
    sut = LgbDataSlicer()
    actual = sut.slice(data, [2, 0])
    assert_array_equal(actual, expected)


def test_pandas_df() -> None:
    data = pd.DataFrame([[1, 2], [3, 4], [5, 6]])
    expected = pd.DataFrame([[5, 6], [1, 2]], index=[2, 0])
    sut = LgbDataSlicer()
    actual = sut.slice(data, [2, 0])
    assert_frame_equal(cast(pd.DataFrame, actual), expected)



from __future__ import annotations

from collections.abc import Sequence
from pathlib import Path

import numpy as np
import scipy
from factrainer.base.dataset import BaseDatasetSlicer, RowIndex

import lightgbm as lgb
from lightgbm.basic import (
    _LGBM_GroupType,
    _LGBM_InitScoreType,
    _LGBM_LabelType,
    _LGBM_PositionType,
    _LGBM_TrainDataType,
    _LGBM_WeightType,
)
from lightgbm.compat import (
    dt_DataTable,
)

from .types import (
    IsPaArray,
    IsPaChunkedArray,
    IsPaTable,
    IsPdDataFrame,
    IsPdSeries,
)


class LgbDataSlicer(BaseDatasetSlicer[_LGBM_TrainDataType]):
    def slice(self, data: _LGBM_TrainDataType, index: RowIndex) -> _LGBM_TrainDataType:
        if isinstance(data, str):
            raise NotImplementedError
        elif isinstance(data, Path):
            raise NotImplementedError
        elif isinstance(data, np.ndarray):
            return data[index]
        elif IsPdDataFrame().is_instance(data):
            return data.take(index)
        elif isinstance(data, dt_DataTable):
            raise NotImplementedError
        elif isinstance(data, scipy.sparse.spmatrix):
            raise NotImplementedError
        elif isinstance(data, Sequence):
            raise NotImplementedError
        elif isinstance(data, list):
            raise NotImplementedError
        elif IsPaTable().is_instance(data):
            raise NotImplementedError
        else:
            raise TypeError


class LgbLabelSlicer(BaseDatasetSlicer[_LGBM_LabelType]):
    def slice(self, data: _LGBM_LabelType, index: RowIndex) -> _LGBM_LabelType:
        if isinstance(data, list):
            raise NotImplementedError
        elif isinstance(data, np.ndarray):
            return data[index]
        elif IsPdDataFrame().is_instance(data):
            return data.take(index)
        elif IsPdSeries().is_instance(data):
            return data.take(index)
        elif isinstance(data, dt_DataTable):
            raise NotImplementedError
        elif IsPaArray().is_instance(data):
            raise NotImplementedError
        elif IsPaChunkedArray().is_instance(data):
            raise NotImplementedError
        else:
            raise TypeError


class LgbWeightSlicer(BaseDatasetSlicer[_LGBM_WeightType]):
    def slice(self, data: _LGBM_WeightType, index: RowIndex) -> _LGBM_WeightType:
        if isinstance(data, list):
            return [data[i] for i in index]
        elif isinstance(data, np.ndarray):
            return data[index]
        elif IsPdSeries().is_instance(data):
            return data.take(index)
        elif IsPaArray().is_instance(data):
            raise NotImplementedError
        elif IsPaChunkedArray().is_instance(data):
            raise NotImplementedError
        else:
            raise TypeError


class LgbInitScoreSlicer(BaseDatasetSlicer[_LGBM_InitScoreType]):
    def slice(self, data: _LGBM_InitScoreType, index: RowIndex) -> _LGBM_InitScoreType:
        if isinstance(data, list):
            raise NotImplementedError
        elif isinstance(data, np.ndarray):
            return data[index]
        elif IsPdDataFrame().is_instance(data):
            return data.take(index)
        elif IsPdSeries().is_instance(data):
            return data.take(index)
        elif IsPaTable().is_instance(data):
            raise NotImplementedError
        elif IsPaArray().is_instance(data):
            raise NotImplementedError
        elif IsPaChunkedArray().is_instance(data):
            raise NotImplementedError
        else:
            raise TypeError


class LgbGroupSlicer(BaseDatasetSlicer[_LGBM_GroupType]):
    def slice(self, data: _LGBM_GroupType, index: RowIndex) -> _LGBM_GroupType:
        if isinstance(data, list):
            raise NotImplementedError
        elif isinstance(data, np.ndarray):
            return data[index]
        elif IsPdSeries().is_instance(data):
            return data.take(index)
        elif IsPaArray().is_instance(data):
            raise NotImplementedError
        elif IsPaChunkedArray().is_instance(data):
            raise NotImplementedError
        else:
            raise TypeError


class LgbPositionSlicer(BaseDatasetSlicer[_LGBM_PositionType]):
    def slice(self, data: _LGBM_PositionType, index: RowIndex) -> _LGBM_PositionType:
        if isinstance(data, np.ndarray):
            return data[index]
        elif IsPdSeries().is_instance(data):
            return data.take(index)
        else:
            raise TypeError


class LgbDatasetSlicer(BaseDatasetSlicer[lgb.Dataset]):
    def __init__(self, reference: lgb.Dataset | None = None) -> None:
        self.reference = reference

    def slice(self, data: lgb.Dataset, index: RowIndex) -> lgb.Dataset:
        return lgb.Dataset(
            data=LgbDataSlicer().slice(data.data, index),
            label=(
                LgbLabelSlicer().slice(data.label, index)
                if data.label is not None
                else None
            ),
            reference=self.reference,
            weight=(
                LgbWeightSlicer().slice(data.weight, index)
                if data.weight is not None
                else None
            ),
            group=(
                LgbGroupSlicer().slice(data.group, index)
                if data.group is not None
                else None
            ),
            init_score=(
                LgbInitScoreSlicer().slice(data.init_score, index)
                if data.init_score is not None
                else None
            ),
            feature_name=data.feature_name,
            categorical_feature=data.categorical_feature,
            params=data.params,
            free_raw_data=data.free_raw_data,
            position=(
                LgbPositionSlicer().slice(data.position, index)
                if data.position is not None
                else None
            ),
        )



from inspect import Parameter, Signature, signature
from pathlib import Path
from typing import Any, Callable, Dict, List, Optional, Union

import lightgbm as lgb
import numpy as np
import scipy
from lightgbm import Dataset
from lightgbm.basic import (
    _LGBM_CategoricalFeatureConfiguration,
    _LGBM_FeatureNameConfiguration,
    _LGBM_GroupType,
    _LGBM_InitScoreType,
    _LGBM_LabelType,
    _LGBM_PositionType,
    _LGBM_PredictDataType,
    _LGBM_TrainDataType,
    _LGBM_WeightType,
)


def test_train_signature() -> None:
    expected = Signature(
        [
            Parameter(
                "params", Parameter.POSITIONAL_OR_KEYWORD, annotation=Dict[str, Any]
            ),
            Parameter(
                "train_set", Parameter.POSITIONAL_OR_KEYWORD, annotation=lgb.Dataset
            ),
            Parameter(
                "num_boost_round",
                Parameter.POSITIONAL_OR_KEYWORD,
                annotation=int,
                default=100,
            ),
            Parameter(
                "valid_sets",
                Parameter.POSITIONAL_OR_KEYWORD,
                annotation=Optional[List[lgb.Dataset]],
                default=None,
            ),
            Parameter(
                "valid_names",
                Parameter.POSITIONAL_OR_KEYWORD,
                annotation=Optional[List[str]],
                default=None,
            ),
            Parameter(
                "feval",
                Parameter.POSITIONAL_OR_KEYWORD,
                annotation=Optional[
                    Union[
                        lgb.engine._LGBM_CustomMetricFunction,
                        List[lgb.engine._LGBM_CustomMetricFunction],
                    ]
                ],
                default=None,
            ),
            Parameter(
                "init_model",
                Parameter.POSITIONAL_OR_KEYWORD,
                annotation=Optional[Union[str, Path, lgb.Booster]],
                default=None,
            ),
            Parameter(
                "keep_training_booster",
                Parameter.POSITIONAL_OR_KEYWORD,
                annotation=bool,
                default=False,
            ),
            Parameter(
                "callbacks",
                Parameter.POSITIONAL_OR_KEYWORD,
                annotation=Optional[List[Callable]],  # type: ignore
                default=None,
            ),
        ],
        return_annotation=lgb.Booster,
    )
    actual = signature(lgb.train)

    assert actual == expected


def test_predict_signature() -> None:
    expected = Signature(
        [
            Parameter("self", Parameter.POSITIONAL_OR_KEYWORD),
            Parameter(
                "data",
                Parameter.POSITIONAL_OR_KEYWORD,
                annotation=_LGBM_PredictDataType,
            ),
            Parameter(
                "start_iteration",
                Parameter.POSITIONAL_OR_KEYWORD,
                annotation=int,
                default=0,
            ),
            Parameter(
                "num_iteration",
                Parameter.POSITIONAL_OR_KEYWORD,
                annotation=Optional[int],
                default=None,
            ),
            Parameter(
                "raw_score",
                Parameter.POSITIONAL_OR_KEYWORD,
                annotation=bool,
                default=False,
            ),
            Parameter(
                "pred_leaf",
                Parameter.POSITIONAL_OR_KEYWORD,
                annotation=bool,
                default=False,
            ),
            Parameter(
                "pred_contrib",
                Parameter.POSITIONAL_OR_KEYWORD,
                annotation=bool,
                default=False,
            ),
            Parameter(
                "data_has_header",
                Parameter.POSITIONAL_OR_KEYWORD,
                annotation=bool,
                default=False,
            ),
            Parameter(
                "validate_features",
                Parameter.POSITIONAL_OR_KEYWORD,
                annotation=bool,
                default=False,
            ),
            Parameter("kwargs", Parameter.VAR_KEYWORD, annotation=Any),
        ],
        return_annotation=Union[
            np.ndarray, scipy.sparse.spmatrix, List[scipy.sparse.spmatrix]
        ],
    )

    actual = signature(lgb.Booster.predict)

    assert actual == expected


def test_dataset_signature() -> None:
    expected = Signature(
        [
            Parameter(
                "data",
                Parameter.POSITIONAL_OR_KEYWORD,
                annotation=_LGBM_TrainDataType,
            ),
            Parameter(
                "label",
                Parameter.POSITIONAL_OR_KEYWORD,
                annotation=Optional[_LGBM_LabelType],
                default=None,
            ),
            Parameter(
                "reference",
                Parameter.POSITIONAL_OR_KEYWORD,
                annotation=Optional["Dataset"],
                default=None,
            ),
            Parameter(
                "weight",
                Parameter.POSITIONAL_OR_KEYWORD,
                annotation=Optional[_LGBM_WeightType],
                default=None,
            ),
            Parameter(
                "group",
                Parameter.POSITIONAL_OR_KEYWORD,
                annotation=Optional[_LGBM_GroupType],
                default=None,
            ),
            Parameter(
                "init_score",
                Parameter.POSITIONAL_OR_KEYWORD,
                annotation=Optional[_LGBM_InitScoreType],
                default=None,
            ),
            Parameter(
                "feature_name",
                Parameter.POSITIONAL_OR_KEYWORD,
                annotation=_LGBM_FeatureNameConfiguration,
                default="auto",
            ),
            Parameter(
                "categorical_feature",
                Parameter.POSITIONAL_OR_KEYWORD,
                annotation=_LGBM_CategoricalFeatureConfiguration,
                default="auto",
            ),
            Parameter(
                "params",
                Parameter.POSITIONAL_OR_KEYWORD,
                annotation=Optional[Dict[str, Any]],
                default=None,
            ),
            Parameter(
                "free_raw_data",
                Parameter.POSITIONAL_OR_KEYWORD,
                annotation=bool,
                default=True,
            ),
            Parameter(
                "position",
                Parameter.POSITIONAL_OR_KEYWORD,
                annotation=Optional[_LGBM_PositionType],
                default=None,
            ),
        ],
    )
    actual = signature(lgb.Dataset)
    assert actual == expected



from __future__ import annotations

from collections.abc import Sequence
from typing import Any, Literal

import numpy as np
import pandas as pd
from numpy import typing as npt

def r2_score(
    y_true: Sequence[Any] | npt.NDArray[np.number[Any]] | pd.DataFrame | pd.Series[Any],
    y_pred: npt.ArrayLike,
    *,
    sample_weight: Sequence[Any] | None = None,
    multioutput: str = "uniform_average",
    force_finite: bool = True,
) -> float: ...
def accuracy_score(
    y_true: Sequence[Any] | npt.NDArray[np.number[Any]] | pd.DataFrame | pd.Series[Any],
    y_pred: npt.ArrayLike,
    *,
    normalize: bool = True,
    sample_weight: Sequence[Any] | None = None,
) -> float: ...
def f1_score(
    y_true: Sequence[Any] | npt.NDArray[np.number[Any]] | pd.DataFrame | pd.Series[Any],
    y_pred: npt.ArrayLike,
    labels: npt.ArrayLike | None = None,
    pos_label: str | int = 1,
    average: Literal["micro", "macro", "samples", "weighted", "binary"]
    | None = "binary",
    sample_weight: npt.ArrayLike | None = None,
    zero_division: int | Literal["warn"] = "warn",
) -> float: ...

#   *,
#     labels: ArrayLike | None = None,
#     pos_label: str | int = 1,
#     average: Literal['micro', 'macro', 'samples', 'weighted', 'binary'] | None = "binary",
#     sample_weight: ArrayLike | None = None,
#     zero_division: int | Literal['warn'] = "warn"
# ) -> (Float | ndarray)



from __future__ import annotations

from typing import cast

import numpy as np
import pandas as pd
import polars as pl
from factrainer.sklearn.dataset.dataset import SklearnDataset
from numpy.testing import assert_array_equal
from pandas.testing import assert_frame_equal, assert_series_equal
from polars.testing import assert_frame_equal as pl_assert_frame_equal
from polars.testing import assert_series_equal as pl_assert_series_equal


class TestSklearnDatasetGetitem:
    def test_single_row(self) -> None:
        sut = SklearnDataset(X=np.array([[1, 2, 3], [4, 5, 6]]), y=np.array([10, 20]))
        expected = SklearnDataset(X=np.array([[4, 5, 6]]), y=np.array([20]))
        actual = sut[1]
        assert_array_equal(actual.X, expected.X)
        assert_array_equal(actual.y, expected.y)

    def test_rows(self) -> None:
        sut = SklearnDataset(
            X=np.array([[1, 2, 3], [4, 5, 6], [7, 8, 9]]), y=np.array([10, 20, 30])
        )
        expected = SklearnDataset(
            X=np.array([[7, 8, 9], [1, 2, 3]]), y=np.array([30, 10])
        )
        actual = sut[[2, 0]]
        assert_array_equal(actual.X, expected.X)
        assert_array_equal(actual.y, expected.y)

    def test_pandas(self) -> None:
        X = pd.DataFrame(
            {
                "a": [1, 2, 3],
                "b": [4, 5, 6],
            }
        )
        y = pd.Series([10, 20, 30])
        sut = SklearnDataset(X=X, y=y)
        expected_X = pd.DataFrame(
            {
                "a": [3, 1],
                "b": [6, 4],
            },
            index=[2, 0],
        )
        expected_y = pd.Series([30, 10], index=[2, 0])
        actual = sut[[2, 0]]
        assert_frame_equal(cast(pd.DataFrame, actual.X), expected_X)
        if not isinstance(actual.y, pd.Series):
            raise TypeError
        assert_series_equal(actual.y, expected_y)

    def test_polars(self) -> None:
        X = pl.DataFrame(
            {
                "a": [1, 2, 3],
                "b": [4, 5, 6],
            }
        )
        y = pl.Series([10, 20, 30])
        sut = SklearnDataset(X=X, y=y)
        expected_X = pl.DataFrame(
            {
                "a": [3, 1],
                "b": [6, 4],
            }
        )
        expected_y = pl.Series([30, 10])
        actual = sut[[2, 0]]
        pl_assert_frame_equal(cast(pl.DataFrame, actual.X), expected_X)
        pl_assert_series_equal(cast(pl.Series, actual.y), expected_y)



from typing import Any, Literal, Optional, Union, overload

from numpy import typing as npt

class Bunch: ...

def fetch_california_housing(
    *,
    data_home: Optional[str] = None,
    download_if_missing: bool = True,
    return_X_y: bool = False,
    as_frame: bool = False,
    n_retries: int = 3,
    delay: float = 1.0,
) -> Bunch: ...
def fetch_openml(
    *,
    name: Optional[str] = None,
    version: Union[str, int] = "active",
    data_id: Optional[int] = None,
    data_home: Optional[str] = None,
    target_column: Optional[Union[str, list[str]]] = "default-target",
    cache: bool = True,
    return_X_y: bool = False,
    as_frame: bool = False,
    parser: str = "auto",
    n_retries: int = 3,
    delay: float = 1.0,
) -> Bunch: ...
@overload
def load_iris(
    *,
    return_X_y: Literal[False],
    as_frame: bool = False,
) -> Bunch: ...
@overload
def load_iris(
    *,
    return_X_y: Literal[True],
    as_frame: bool = False,
) -> tuple[npt.NDArray[Any], npt.NDArray[Any]]: ...
@overload
def load_iris(
    *,
    return_X_y: bool = False,
    as_frame: bool = False,
) -> Bunch | tuple[npt.NDArray[Any], npt.NDArray[Any]]: ...
@overload
def load_breast_cancer(
    *,
    return_X_y: Literal[False],
    as_frame: bool = False,
) -> Bunch: ...
@overload
def load_breast_cancer(
    *,
    return_X_y: Literal[True],
    as_frame: bool = False,
) -> tuple[npt.NDArray[Any], npt.NDArray[Any]]: ...
@overload
def load_breast_cancer(
    *,
    return_X_y: bool = False,
    as_frame: bool = False,
) -> Bunch | tuple[npt.NDArray[Any], npt.NDArray[Any]]: ...



from __future__ import annotations

import tempfile
from typing import Any

import joblib
import numpy as np
import pytest
from factrainer.core import CvModelContainer, PredMode
from factrainer.sklearn import (
    SklearnDataset,
    SklearnModel,
    SklearnModelConfig,
    SklearnPredictConfig,
    SklearnPredictMethod,
    SklearnTrainConfig,
)
from numpy import typing as npt
from numpy.testing import assert_allclose
from sklearn.ensemble import RandomForestRegressor
from sklearn.linear_model import LogisticRegression
from sklearn.metrics import accuracy_score, f1_score, r2_score
from sklearn.model_selection import KFold, train_test_split
from sklearn.pipeline import Pipeline
from sklearn.preprocessing import StandardScaler
from sklearn.svm import SVC


class TestClassification:
    @pytest.mark.flaky(reruns=3, reruns_delay=5, only_rerun=["HTTPError"])
    def test_preidct_auto(
        self,
        iris_data: tuple[npt.NDArray[np.number[Any]], npt.NDArray[np.number[Any]]],
    ) -> None:
        features, target = iris_data
        dataset = SklearnDataset(X=features, y=target)
        config = SklearnModelConfig.create(
            train_config=SklearnTrainConfig(estimator=LogisticRegression()),
        )
        k_fold = KFold(n_splits=4, shuffle=True, random_state=1)
        model = CvModelContainer(config, k_fold)
        model.train(dataset)
        y_pred = model.predict(dataset)
        y_pred = np.argmax(y_pred, axis=1)
        metric = f1_score(target, y_pred, average="micro")

        assert_allclose(metric, 0.95, atol=2.5e-02)

    def test_predict(
        self,
        breast_cancer_data: tuple[
            npt.NDArray[np.number[Any]], npt.NDArray[np.number[Any]]
        ],
    ) -> None:
        features, target = breast_cancer_data
        dataset = SklearnDataset(X=features, y=target)
        config = SklearnModelConfig.create(
            train_config=SklearnTrainConfig(estimator=SVC(kernel="linear")),
            pred_config=SklearnPredictConfig(
                predict_method=SklearnPredictMethod.PREDICT
            ),
        )
        k_fold = KFold(n_splits=4, shuffle=True, random_state=1)
        model = CvModelContainer(config, k_fold)
        model.train(dataset)
        y_pred = model.predict(dataset)
        metric = accuracy_score(target, y_pred > 0)

        assert_allclose(metric, 0.95, atol=2.5e-02)

    @pytest.mark.flaky(reruns=3, reruns_delay=5, only_rerun=["HTTPError"])
    def test_preidct_proba(
        self,
        iris_data: tuple[npt.NDArray[np.number[Any]], npt.NDArray[np.number[Any]]],
    ) -> None:
        features, target = iris_data
        dataset = SklearnDataset(X=features, y=target)
        config = SklearnModelConfig.create(
            train_config=SklearnTrainConfig(estimator=LogisticRegression()),
            pred_config=SklearnPredictConfig(
                predict_method=SklearnPredictMethod.PREDICT_PROBA
            ),
        )
        k_fold = KFold(n_splits=4, shuffle=True, random_state=1)
        model = CvModelContainer(config, k_fold)
        model.train(dataset)
        y_pred = model.predict(dataset)
        y_pred = np.argmax(y_pred, axis=1)
        metric = f1_score(target, y_pred, average="micro")

        assert_allclose(metric, 0.95, atol=2.5e-02)


@pytest.mark.flaky(reruns=3, reruns_delay=5, only_rerun=["HTTPError"])
def test_cv_set_pred_config_after_trainig(
    breast_cancer_data: tuple[npt.NDArray[np.number[Any]], npt.NDArray[np.number[Any]]],
) -> None:
    features, target = breast_cancer_data
    dataset = SklearnDataset(X=features, y=target)
    config = SklearnModelConfig.create(
        train_config=SklearnTrainConfig(
            estimator=Pipeline(
                [("scaler", StandardScaler()), ("lr", LogisticRegression())]
            )
        ),
    )
    k_fold = KFold(n_splits=4, shuffle=True, random_state=1)
    model = CvModelContainer(config, k_fold)
    model.train(dataset)
    y_pred_proba = model.predict(dataset)
    model.pred_config = SklearnPredictConfig(
        predict_method=SklearnPredictMethod.PREDICT
    )
    y_pred = model.predict(dataset)

    assert y_pred_proba.ndim == 2
    assert y_pred.ndim == 1


@pytest.mark.flaky(reruns=3, reruns_delay=5, only_rerun=["HTTPError"])
def test_cv_average_ensembling(
    california_housing_data: tuple[
        npt.NDArray[np.number[Any]], npt.NDArray[np.number[Any]]
    ],
) -> None:
    features, target = california_housing_data
    train_X, test_X, train_y, test_y = train_test_split(
        features,
        target,
        test_size=0.2,
        random_state=1,
    )
    train_dataset = SklearnDataset(X=train_X, y=train_y)
    test_dataset = SklearnDataset(X=test_X, y=test_y)
    config = SklearnModelConfig.create(
        train_config=SklearnTrainConfig(
            estimator=RandomForestRegressor(random_state=100, n_jobs=-1)
        ),
    )
    k_fold = KFold(n_splits=5, shuffle=True, random_state=1)
    model = CvModelContainer(config, k_fold)
    model.train(train_dataset)
    y_pred = model.predict(test_dataset, n_jobs=4, mode=PredMode.AVG_ENSEMBLE)
    metric = r2_score(test_y, y_pred)

    assert_allclose(metric, 0.8, atol=2.5e-02)


@pytest.mark.flaky(reruns=3, reruns_delay=5, only_rerun=["HTTPError"])
def test_cv_model_picklable(
    iris_data: tuple[npt.NDArray[np.number[Any]], npt.NDArray[np.number[Any]]],
) -> None:
    features, target = iris_data
    dataset = SklearnDataset(X=features, y=target)
    config = SklearnModelConfig.create(
        train_config=SklearnTrainConfig(estimator=LogisticRegression()),
    )
    k_fold = KFold(n_splits=4, shuffle=True, random_state=1)
    model = CvModelContainer(config, k_fold)
    model.train(dataset)
    with tempfile.NamedTemporaryFile() as fp:
        joblib.dump(model, fp.name)
        loaded_model: CvModelContainer[
            SklearnDataset, SklearnModel, SklearnTrainConfig, SklearnPredictConfig
        ] = joblib.load(fp.name)
    y_pred = loaded_model.predict(dataset)
    y_pred = np.argmax(y_pred, axis=1)
    metric = f1_score(target, y_pred, average="micro")

    assert_allclose(metric, 0.95, atol=2.5e-02)



from __future__ import annotations

from typing import TYPE_CHECKING, Any, Protocol, runtime_checkable

from factrainer.base.dataset import IsImportableInstance
from typing_extensions import TypeIs

if TYPE_CHECKING:
    import pandas as pd
    import pyarrow as pa


@runtime_checkable
class PdDataFrameProtocol(Protocol): ...


@runtime_checkable
class PdSeriesProtocol[T](Protocol): ...


@runtime_checkable
class PaTableProtocol(Protocol): ...


@runtime_checkable
class PaArrayProtocol(Protocol): ...


@runtime_checkable
class PaChunkedArrayProtocol(Protocol): ...


if TYPE_CHECKING:
    type pd_DataFrame = pd.DataFrame
    type pd_Series[T] = pd.Series[T]
    type pa_Table = pa.Table
    type pa_Array = pa.Array[Any]
    type pa_ChunkedArray = pa.ChunkedArray[Any]
else:
    type pd_DataFrame = PdDataFrameProtocol
    type pd_Series[T] = PdSeriesProtocol[T]
    type pa_Table = PaTableProtocol
    type pa_Array = PaArrayProtocol
    type pa_ChunkedArray = PaChunkedArrayProtocol


class IsPdDataFrame(IsImportableInstance[pd_DataFrame]):
    def is_instance(self, obj: Any) -> TypeIs[pd_DataFrame]:
        try:
            import pandas as pd

            return isinstance(obj, pd.DataFrame)
        except ImportError:
            return False


class IsPdSeries(IsImportableInstance[pd_Series[Any]]):
    def is_instance(self, obj: Any) -> TypeIs[pd_Series[Any]]:
        try:
            import pandas as pd

            return isinstance(obj, pd.Series)
        except ImportError:
            return False


class IsPaTable(IsImportableInstance[pa_Table]):
    def is_instance(self, obj: Any) -> TypeIs[pa_Table]:
        try:
            import pyarrow as pa

            return isinstance(obj, pa.Table)
        except ImportError:
            return False


class IsPaArray(IsImportableInstance[pa_Array]):
    def is_instance(self, obj: Any) -> TypeIs[pa_Array]:
        try:
            import pyarrow as pa

            return isinstance(obj, pa.Array)
        except ImportError:
            return False


class IsPaChunkedArray(IsImportableInstance[pa_ChunkedArray]):
    def is_instance(self, obj: Any) -> TypeIs[pa_ChunkedArray]:
        try:
            import pyarrow as pa

            return isinstance(obj, pa.ChunkedArray)
        except ImportError:
            return False



from __future__ import annotations

from typing import Any

import numpy as np
import pandas as pd
import polars as pl
import pytest
from factrainer.core import (
    CvModelContainer,
)
from factrainer.sklearn import (
    SklearnDataset,
    SklearnModelConfig,
    SklearnTrainConfig,
)
from numpy import typing as npt
from numpy.testing import assert_allclose
from sklearn.linear_model import LogisticRegression
from sklearn.metrics import f1_score
from sklearn.model_selection import KFold


@pytest.mark.flaky(reruns=3, reruns_delay=5, only_rerun=["HTTPError"])
def test_cv_pandas(
    iris_data: tuple[npt.NDArray[np.number[Any]], npt.NDArray[np.number[Any]]],
) -> None:
    features, target = iris_data
    dataset = SklearnDataset(X=pd.DataFrame(features), y=pd.Series(target))
    config = SklearnModelConfig.create(
        train_config=SklearnTrainConfig(estimator=LogisticRegression())
    )

    k_fold = KFold(n_splits=4, shuffle=True, random_state=1)
    model = CvModelContainer(config, k_fold)
    model.train(dataset, n_jobs=4)
    y_pred = model.predict(dataset)
    y_pred = np.argmax(y_pred, axis=1)
    metric = f1_score(target, y_pred, average="micro")

    assert_allclose(metric, 0.95, atol=2.5e-02)


@pytest.mark.flaky(reruns=3, reruns_delay=5, only_rerun=["HTTPError"])
def test_cv_polars(
    iris_data: tuple[npt.NDArray[np.number[Any]], npt.NDArray[np.number[Any]]],
) -> None:
    features, target = iris_data
    dataset = SklearnDataset(X=pl.DataFrame(features), y=pl.Series(target))
    config = SklearnModelConfig.create(
        train_config=SklearnTrainConfig(estimator=LogisticRegression())
    )

    k_fold = KFold(n_splits=4, shuffle=True, random_state=1)
    model = CvModelContainer(config, k_fold)
    model.train(dataset, n_jobs=4)
    y_pred = model.predict(dataset)
    y_pred = np.argmax(y_pred, axis=1)
    metric = f1_score(target, y_pred, average="micro")

    assert_allclose(metric, 0.95, atol=2.5e-02)



from collections.abc import Sequence
from typing import Self

from factrainer.base.dataset import (
    BaseDataset,
    IndexableDataset,
    RowIndex,
    RowIndices,
)
from sklearn.model_selection._split import _BaseKFold


class IndexedDataset[T: IndexableDataset](BaseDataset):
    index: RowIndex
    data: T

    def __len__(self) -> int:
        return len(self.index)


class IndexedDatasets[T: IndexableDataset](BaseDataset):
    datasets: Sequence[IndexedDataset[T]]

    def __len__(self) -> int:
        return sum([len(dataset) for dataset in self.datasets])

    @classmethod
    def create(cls, dataset: T, k_fold: _BaseKFold | RowIndices) -> Self:
        if isinstance(k_fold, _BaseKFold):
            raise NotImplementedError
        return cls(
            datasets=[
                IndexedDataset(index=index, data=dataset[index]) for index in k_fold
            ]
        )

    @property
    def indices(self) -> RowIndices:
        return [dataset.index for dataset in self.datasets]


class SplittedDataset[T: IndexableDataset](BaseDataset):
    train: IndexedDataset[T]
    val: IndexedDataset[T] | None
    test: IndexedDataset[T]


class SplittedDatasetsIndices(BaseDataset):
    train: RowIndices
    val: RowIndices | None
    test: RowIndices


class SplittedDatasets[T: IndexableDataset](BaseDataset):
    datasets: Sequence[SplittedDataset[T]]

    @property
    def train(self) -> IndexedDatasets[T]:
        return IndexedDatasets(datasets=[dataset.train for dataset in self.datasets])

    @property
    def val(self) -> IndexedDatasets[T] | None:
        vals = []
        for dataset in self.datasets:
            if dataset.val is None:
                return None
            vals.append(dataset.val)
        return IndexedDatasets(datasets=vals)

    @property
    def test(self) -> IndexedDatasets[T]:
        return IndexedDatasets(datasets=[dataset.test for dataset in self.datasets])

    @property
    def indices(self) -> SplittedDatasetsIndices:
        return SplittedDatasetsIndices(
            train=self.train.indices,
            val=self.val.indices if self.val is not None else None,
            test=self.test.indices,
        )

    @classmethod
    def create(cls, dataset: T, k_fold: _BaseKFold | SplittedDatasetsIndices) -> Self:
        datasets: list[SplittedDataset[T]] = []
        if isinstance(k_fold, SplittedDatasetsIndices):
            if k_fold.val is None:
                for train_index, test_index in zip(k_fold.train, k_fold.test):
                    datasets.append(
                        SplittedDataset(
                            train=IndexedDataset(
                                index=train_index, data=dataset[train_index]
                            ),
                            val=None,
                            test=IndexedDataset(
                                index=test_index, data=dataset[test_index]
                            ),
                        )
                    )
            else:
                for train_index, val_index, test_index in zip(
                    k_fold.train, k_fold.val, k_fold.test
                ):
                    datasets.append(
                        SplittedDataset(
                            train=IndexedDataset(
                                index=train_index, data=dataset[train_index]
                            ),
                            val=IndexedDataset(
                                index=val_index, data=dataset[val_index]
                            ),
                            test=IndexedDataset(
                                index=test_index, data=dataset[test_index]
                            ),
                        )
                    )
        else:
            for train_index, test_index in dataset.k_fold_split(k_fold):
                datasets.append(
                    SplittedDataset(
                        train=IndexedDataset(
                            index=train_index, data=dataset[train_index]
                        ),
                        val=IndexedDataset(index=test_index, data=dataset[test_index]),
                        test=IndexedDataset(index=test_index, data=dataset[test_index]),
                    )
                )
        return cls(datasets=datasets)



from factrainer.base.config import BaseMlModelConfig, BasePredictConfig, BaseTrainConfig
from factrainer.base.dataset import IndexableDataset, Prediction
from factrainer.base.raw_model import RawModel
from sklearn.model_selection._split import _BaseKFold

from ..model_container import BaseModelContainer
from .config import AverageEnsemblePredictor, CvLearner, OutOfFoldPredictor, PredMode
from .dataset import IndexedDatasets, SplittedDatasets, SplittedDatasetsIndices
from .raw_model import RawModels


class CvModelContainer[
    T: IndexableDataset,
    U: RawModel,
    V: BaseTrainConfig,
    W: BasePredictConfig,
](BaseModelContainer[T, RawModels[U], V, W]):
    def __init__(
        self,
        model_config: BaseMlModelConfig[T, U, V, W],
        k_fold: _BaseKFold | SplittedDatasetsIndices,
    ) -> None:
        self._learner = model_config.learner
        self._predictor = model_config.predictor
        self._train_config = model_config.train_config
        self._pred_config = model_config.pred_config
        self._k_fold = k_fold

    def train(self, train_dataset: T, n_jobs: int | None = None) -> None:
        datasets = SplittedDatasets.create(train_dataset, self._k_fold)
        self._cv_indices = datasets.indices
        self._model = CvLearner(self._learner).train(
            datasets.train, datasets.val, self.train_config, n_jobs
        )

    def predict(
        self,
        pred_dataset: T,
        n_jobs: int | None = None,
        mode: PredMode = PredMode.OOF_PRED,
    ) -> Prediction:
        match mode:
            case PredMode.OOF_PRED:
                datasets = IndexedDatasets[T].create(pred_dataset, self.cv_indices.test)
                return OutOfFoldPredictor(self._predictor).predict(
                    datasets, self.raw_model, self.pred_config, n_jobs
                )
            case PredMode.AVG_ENSEMBLE:
                return AverageEnsemblePredictor(self._predictor).predict(
                    pred_dataset, self.raw_model, self.pred_config, n_jobs
                )
            case _:
                raise ValueError(f"Invalid prediction mode: {mode}")

    @property
    def raw_model(self) -> RawModels[U]:
        return self._model

    @property
    def train_config(self) -> V:
        return self._train_config

    @train_config.setter
    def train_config(self, config: V) -> None:
        self._train_config = config

    @property
    def pred_config(self) -> W:
        return self._pred_config

    @pred_config.setter
    def pred_config(self, config: W) -> None:
        self._pred_config = config

    @property
    def cv_indices(self) -> SplittedDatasetsIndices:
        return self._cv_indices

    @property
    def k_fold(self) -> _BaseKFold | SplittedDatasetsIndices:
        return self._k_fold



from __future__ import annotations

from typing import Any, cast

import lightgbm as lgb
import numpy as np
import pandas as pd
from factrainer.lightgbm.dataset.dataset import LgbDataset
from numpy import typing as npt
from numpy.testing import assert_array_equal
from pandas.testing import assert_frame_equal, assert_series_equal


class TestLgbDatasetGetitem:
    def test_single_row(self) -> None:
        sut = LgbDataset(dataset=lgb.Dataset(data=np.array([[1, 2, 3], [4, 5, 6]])))
        expected = np.array([[1, 2, 3]])
        actual = sut[0]
        assert_array_equal(cast(npt.NDArray[Any], actual.dataset.data), expected)

    def test_rows(self) -> None:
        sut = LgbDataset(
            dataset=lgb.Dataset(data=np.array([[1, 2, 3], [4, 5, 6], [7, 8, 9]]))
        )
        expected = np.array([[7, 8, 9], [1, 2, 3]])
        actual = sut[[2, 0]]
        assert_array_equal(cast(npt.NDArray[Any], actual.dataset.data), expected)

    def test_pandas(self) -> None:
        sut = LgbDataset(
            dataset=lgb.Dataset(
                data=pd.DataFrame(
                    {
                        "a": [1, 2, 3],
                        "b": ["4", pd.NA, "6"],
                    }
                )
            )
        )
        expected = pd.DataFrame(
            {
                "a": [2, 1],
                "b": [pd.NA, "4"],
            },
            index=[1, 0],
        )

        actual = sut[[1, 0]]
        assert_frame_equal(cast(pd.DataFrame, actual.dataset.data), expected)

    def test_labels(self) -> None:
        sut = LgbDataset(
            dataset=lgb.Dataset(
                data=np.array([[1, 2, 3], [4, 5, 6], [7, 8, 9]]),
                label=np.array([10, 20, 30]),
            )
        )
        expected_data = np.array([[7, 8, 9], [1, 2, 3]])
        expected_labels = np.array([30, 10])

        actual = sut[[2, 0]]

        assert_array_equal(cast(npt.NDArray[Any], actual.dataset.data), expected_data)
        assert_array_equal(
            cast(npt.NDArray[Any], actual.dataset.label), expected_labels
        )

    def test_labels_pandas(self) -> None:
        sut = LgbDataset(
            dataset=lgb.Dataset(
                data=pd.DataFrame(
                    {
                        "a": [1, 2, 3],
                        "b": ["4", pd.NA, "6"],
                    }
                ),
                label=pd.Series([10, 20, 30]),
            )
        )
        expected_data = pd.DataFrame(
            {
                "a": [2, 1],
                "b": [pd.NA, "4"],
            },
            index=[1, 0],
        )
        expected_labels = pd.Series([20, 10], index=[1, 0])

        actual = sut[[1, 0]]

        assert_frame_equal(cast(pd.DataFrame, actual.dataset.data), expected_data)
        if not isinstance(actual.dataset.label, pd.Series):
            raise TypeError
        assert_series_equal(actual.dataset.label, expected_labels)

    def test_fixed_attributes(self) -> None:
        sut = LgbDataset(
            dataset=lgb.Dataset(
                data=np.array([[1, 2, 3], [4, 5, 6], [7, 8, 9]]),
                feature_name=["a", "b", "c"],
                categorical_feature=["a"],
                params={"foo": "bar"},
                free_raw_data=False,
            )
        )
        expected_data = np.array([[7, 8, 9], [1, 2, 3]])

        actual = sut[[2, 0]]

        assert_array_equal(cast(npt.NDArray[Any], actual.dataset.data), expected_data)
        assert actual.dataset.feature_name == ["a", "b", "c"]
        assert actual.dataset.categorical_feature == ["a"]
        assert actual.dataset.params == {"foo": "bar"}
        assert actual.dataset.free_raw_data is False



from __future__ import annotations

import lightgbm as lgb
import pandas as pd
import pytest
from factrainer.core import (
    CvModelContainer,
)
from factrainer.lightgbm import (
    LgbDataset,
    LgbModelConfig,
    LgbTrainConfig,
)
from numpy.testing import assert_allclose
from sklearn.metrics import accuracy_score
from sklearn.model_selection import KFold


@pytest.mark.flaky(reruns=3, reruns_delay=5, only_rerun=["HTTPError"])
def test_cv_pandas(titanic_data: tuple[pd.DataFrame, pd.Series[int]]) -> None:
    features, target = titanic_data
    dataset = LgbDataset(dataset=lgb.Dataset(features, label=target))
    config = LgbModelConfig.create(
        train_config=LgbTrainConfig(
            params={"objective": "binary", "seed": 1, "deterministic": True},
            callbacks=[lgb.early_stopping(100, verbose=False)],
        ),
    )
    k_fold = KFold(n_splits=4, shuffle=True, random_state=1)
    model = CvModelContainer(config, k_fold)
    model.train(dataset, n_jobs=4)
    y_pred = model.predict(dataset, n_jobs=4)
    metric = accuracy_score(target, y_pred > 0.5)

    assert_allclose(metric, 0.81, atol=2.5e-02)



from __future__ import annotations

from typing import Any, cast

import numpy as np
import pytest
from factrainer.core import CvModelContainer, SingleModelContainer
from factrainer.sklearn import (
    SklearnDataset,
    SklearnModelConfig,
    SklearnTrainConfig,
)
from factrainer.sklearn.raw_model import Predictable
from numpy import typing as npt
from numpy.testing import assert_allclose
from sklearn.ensemble import RandomForestRegressor
from sklearn.exceptions import NotFittedError
from sklearn.metrics import r2_score
from sklearn.model_selection import KFold, train_test_split


@pytest.mark.flaky(reruns=3, reruns_delay=5, only_rerun=["HTTPError"])
def test_cv_model_regression(
    california_housing_data: tuple[
        npt.NDArray[np.number[Any]], npt.NDArray[np.number[Any]]
    ],
) -> None:
    features, target = california_housing_data
    dataset = SklearnDataset(X=features, y=target)
    config = SklearnModelConfig.create(
        train_config=SklearnTrainConfig(
            estimator=RandomForestRegressor(random_state=100, n_jobs=-1)
        ),
    )
    k_fold = KFold(n_splits=4, shuffle=True, random_state=1)
    model = CvModelContainer(config, k_fold)
    model.train(dataset)
    y_pred = model.predict(dataset)
    metric = r2_score(target, y_pred)

    assert_allclose(metric, 0.8, atol=2.5e-02)
    with pytest.raises(NotFittedError):
        cast(Predictable, config.train_config.estimator).predict(np.array([]))


@pytest.mark.flaky(reruns=3, reruns_delay=5, only_rerun=["HTTPError"])
def test_cv_model_parallel(
    california_housing_data: tuple[
        npt.NDArray[np.number[Any]], npt.NDArray[np.number[Any]]
    ],
) -> None:
    features, target = california_housing_data
    dataset = SklearnDataset(X=features, y=target)
    config = SklearnModelConfig.create(
        train_config=SklearnTrainConfig(
            estimator=RandomForestRegressor(random_state=100, n_jobs=-1)
        ),
    )
    k_fold = KFold(n_splits=4, shuffle=True, random_state=1)
    model = CvModelContainer(config, k_fold)
    model.train(dataset, n_jobs=2)
    y_pred = model.predict(dataset)
    metric = r2_score(target, y_pred)

    assert_allclose(metric, 0.8, atol=2.5e-02)


@pytest.mark.flaky(reruns=3, reruns_delay=5, only_rerun=["HTTPError"])
def test_single_model(
    california_housing_data: tuple[
        npt.NDArray[np.number[Any]], npt.NDArray[np.number[Any]]
    ],
) -> None:
    features, target = california_housing_data
    train_X, test_X, train_y, test_y = train_test_split(
        features,
        target,
        test_size=0.2,
        random_state=1,
    )
    train_dataset = SklearnDataset(X=train_X, y=train_y)
    test_dataset = SklearnDataset(X=test_X, y=test_y)
    config = SklearnModelConfig.create(
        train_config=SklearnTrainConfig(
            estimator=RandomForestRegressor(random_state=100, n_jobs=-1)
        ),
    )
    model = SingleModelContainer(config)
    model.train(train_dataset)
    y_pred = model.predict(test_dataset)
    metric = r2_score(test_y, y_pred)

    assert_allclose(metric, 0.8, atol=2.5e-02)



from __future__ import annotations

from typing import Any, cast

import numpy as np
import pandas as pd
from factrainer.lightgbm.dataset.slicer import LgbWeightSlicer
from numpy.testing import assert_array_equal
from pandas.testing import assert_series_equal


def test_numpy_1d_array() -> None:
    data = np.array([1, 2, 3, 4, 5, 6])
    expected = np.array([3, 1])
    sut = LgbWeightSlicer()
    actual = sut.slice(data, [2, 0])
    assert_array_equal(actual, expected)


def test_pandas_series() -> None:
    data = pd.Series([1, 2, 3, 4, 5, 6])
    expected = pd.Series([3, 1], index=[2, 0])
    sut = LgbWeightSlicer()
    actual = sut.slice(data, [2, 0])
    assert_series_equal(cast("pd.Series[Any]", actual), expected)



from .config import LgbModelConfig, LgbPredictConfig, LgbTrainConfig
from .dataset.dataset import LgbDataset
from .raw_model import LgbModel

__all__ = [
    "LgbDataset",
    "LgbModelConfig",
    "LgbPredictConfig",
    "LgbTrainConfig",
    "LgbModel",
]
